searchNodes=[{"doc":"TypedStruct is a library for defining structs with a type without writing boilerplate code.RationaleTo define a struct in Elixir, you probably want to define three things:the struct itself, with default values,the list of enforced keys,its associated type.It ends up in something like this:defmodule Person do @moduledoc &quot;&quot;&quot; A struct representing a person. &quot;&quot;&quot; @enforce_keys [:name] defstruct name: nil, age: nil, happy?: true, phone: nil @typedoc &quot;A person&quot; @type t() :: %__MODULE__{ name: String.t(), age: non_neg_integer() | nil, happy?: boolean(), phone: String.t() | nil } endIn the example above you can notice several points:the keys are present in both the defstruct and type definition,enforced keys must also be written in @enforce_keys,if a key has no default value and is not enforced, its type should be nullable.If you want to add a field in the struct, you must therefore:add the key with its default value in the defstruct list,add the key with its type in the type definition.If the field is not optional, you should even add it to @enforce_keys. This is way too much work for lazy people like me, and moreover it can be error-prone.It would be way better if we could write something like this:defmodule Person do @moduledoc &quot;&quot;&quot; A struct representing a person. &quot;&quot;&quot; use TypedStruct @typedoc &quot;A person&quot; typedstruct do field :name, String.t(), enforce: true field :age, non_neg_integer() field :happy?, boolean(), default: true field :phone, String.t() end endThanks to TypedStruct, this is now possible :)UsageSetupTo use TypedStruct in your project, add this to your Mix dependencies:{:typed_struct, &quot;~&gt; 0.2.0&quot;}If you do not plan to compile modules using TypedStruct at runtime, you can add runtime: false to the dependency tuple as TypedStruct is only used at build time.If you want to avoid mix format putting parentheses on field definitions, you can add to your .formatter.exs:[ ..., import_deps: [:typed_struct] ]General usageTo define a typed struct, use TypedStruct, then define your struct within a typedstruct block:defmodule MyStruct do # Use TypedStruct to import the typedstruct macro. use TypedStruct # Define your struct. typedstruct do # Define each field with the field macro. field :a_string, String.t() # You can set a default value. field :string_with_default, String.t(), default: &quot;default&quot; # You can enforce a field. field :enforced_field, integer(), enforce: true end endEach field is defined through the field/2 macro.OptionsIf you want to enforce all the keys by default, you can do:defmodule MyStruct do use TypedStruct # Enforce keys by default. typedstruct enforce: true do # This key is enforced. field :enforced_by_default, term() # You can override the default behaviour. field :not_enforced, term(), enforce: false # A key with a default value is not enforced. field :not_enforced_either, integer(), default: 1 end endYou can also generate an opaque type for the struct:defmodule MyOpaqueStruct do use TypedStruct # Generate an opaque type for the struct. typedstruct opaque: true do field :name, String.t() end endDocumentationTo add a @typedoc to the struct type, just add the attribute above the typedstruct block:@typedoc &quot;A typed struct&quot; typedstruct do field :a_string, String.t() field :an_int, integer() endPluginsIt is possible to extend the scope of TypedStruct by using its plugin interface, as described in TypedStruct.Plugin. For instance, to automatically generate lenses with the Lens library, you can use TypedStructLens and do:defmodule MyStruct do use TypedStruct typedstruct do plugin TypedStructLens field :a_field, String.t() field :other_field, atom() end @spec change(t()) :: t() def change(data) do # a_field/0 is generated by TypedStructLens. lens = a_field() put_in(data, [lens], &quot;Changed&quot;) end endWhat do I get?When defining an empty typedstruct block:defmodule Example do use TypedStruct typedstruct do end endyou get an empty struct with its module type t():defmodule Example do @enforce_keys [] defstruct [] @type t() :: %__MODULE__{} endEach field call adds information to the struct, @enforce_keys and the type t().A field with no options adds the name to the defstruct list, with nil as default. The type itself is made nullable:defmodule Example do use TypedStruct typedstruct do field :name, String.t() end endbecomes:defmodule Example do @enforce_keys [] defstruct name: nil @type t() :: %__MODULE__{ name: String.t() | nil } endThe default option adds the default value to the defstruct:field :name, String.t(), default: &quot;John Smith&quot; # Becomes defstruct name: &quot;John Smith&quot;When set to true, the enforce option enforces the key by adding it to the @enforce_keys attribute.field :name, String.t(), enforce: true # Becomes @enforce_keys [:name] defstruct name: nilIn both cases, the type has no reason to be nullable anymore by default. In one case the field is filled with its default value and not nil, and in the other case it is enforced. Both options would generate the following type:@type t() :: %__MODULE__{ name: String.t() # Not nullable }Passing opaque: true replaces @type with @opaque in the struct type specification:typedstruct opaque: true do field :name, String.t() end # Becomes @opaque t() :: %__MODULE__{ name: String.t() }","ref":"TypedStruct.html","title":"TypedStruct","type":"module"},{"doc":"Defines a field in a typed struct.Example# A field named :example of type String.t() field :example, String.t()Optionsdefault - sets the default value for the fieldenforce - if set to true, enforces the field and makes its type non-nullable","ref":"TypedStruct.html#field/3","title":"TypedStruct.field/3","type":"macro"},{"doc":"Registers a plugin for the currently defined struct.Exampletypedstruct do plugin MyPlugin field :a_field, String.t() endFor more information on how to define your own plugins, please see TypedStruct.Plugin. To use a third-party plugin, please refer directly to its documentation.","ref":"TypedStruct.html#plugin/2","title":"TypedStruct.plugin/2","type":"macro"},{"doc":"Defines a typed struct.Inside a typedstruct block, each field is defined through the field/2 macro.Optionsenforce - if set to true, sets enforce: true to all fields by default. This can be overridden by setting enforce: false or a default value on individual fields.opaque - if set to true, creates an opaque type for the structExamplesdefmodule MyStruct do use TypedStruct typedstruct do field :field_one, String.t() field :field_two, integer(), enforce: true field :field_three, boolean(), enforce: true field :field_four, atom(), default: :hey end endThe following is an equivalent using the enforce by default behaviour:defmodule MyStruct do use TypedStruct typedstruct enforce: true do field :field_one, String.t(), enforce: false field :field_two, integer() field :field_three, boolean() field :field_four, atom(), default: :hey end end","ref":"TypedStruct.html#typedstruct/2","title":"TypedStruct.typedstruct/2","type":"macro"},{"doc":"This module defines the plugin interface for TypedStruct.RationaleSometimes you may want to define helpers on your structs, for all their fields or for the struct as a whole. This plugin interface lets you integrate your own needs with TypedStruct.Plugin definitionA TypedStruct plugin is a module that implements TypedStruct.Plugin. Three callbacks can be used to inject code at different steps:init/1 lets you inject code where the TypedStruct.plugin/2 macro is called,field/3 lets you inject code on each field definition,after_definition/1 lets you insert code after the struct and its type have been defined.To provide the maximum flexibility, the last two callbacks are optional and a default implementation of init/1 is defined when use-ing the module, so that you do not have to write it yourself if you do not need it.ExampleAs an example, let’s define a plugin that allows users to add an optional description to their structs and fields. This plugin also takes an upcase option. If set to true, all the descriptions are then upcased. It would be used this way:defmodule MyStruct do use TypedStruct typedstruct do # We import the plugin with the upcase option set to `true`. plugin DescribedStruct, upcase: true # We can now set a description for the struct. description &quot;My struct&quot; # We can also set a description on a field. field :a_field, String.t(), description: &quot;A field&quot; field :second_field, boolean() end endOnce compiled, we would optain:iex&gt; MyStruct.struct_description() &quot;MY STRUCT&quot; iex&gt; MyStruct.field_description(:a_field) &quot;A FIELD&quot;Follows the plugin definition:defmodule DescribedStruct do use TypedStruct.Plugin # The init macro lets you inject code where the plugin macro is called. # You can think a bit of it like a `use` but for the scope of the # typedstruct block. @impl true @spec init(keyword()) :: Macro.t() defmacro init(opts) do quote do # Let’s import our custom `description` macro defined below so our # users can use it when defining their structs. import TypedStructDemoPlugin, only: [description: 1] # Let’s also store the upcase option in an attribute so we can # access it from the code injected by our `description/1` macro. @upcase unquote(opts)[:upcase] end end # This is a public macro our users can call in their typedstruct blocks. @spec description(String.t()) :: Macro.t() defmacro description(description) do quote do # Here we simply evaluate the result of __description__/2. We need # this indirection to be able to use @upcase after is has been # evaluated, but still in the code generation process. This way, we # can upcase the strings *at build time* if needed. It’s just a tiny # refinement :-) Module.eval_quoted( __MODULE__, TypedStructDemoPlugin.__description__(__MODULE__, unquote(description)) ) end end @spec __description__(module(), String.t()) :: Macro.t() def __description__(module, description) do # Maybe upcase the description at build time. description = module |&gt; Module.get_attribute(:upcase) |&gt; maybe_upcase(description) quote do # Let’s just generate a constant function that returns the # description. def struct_description, do: unquote(description) end end # The field callback is called for each field defined in the typedstruct # block. You get exactly what the user has passed to the field macro, # plus options from every plugin init. @impl true @spec field(atom(), any(), keyword()) :: Macro.t() def field(name, _type, opts) do # Same as for the struct description, we want to upcase at build time # if necessary. As we do not have access to the module here, we cannot # access @upcase. This is not an issue since the option is # automatically added to `opts`, in addition to the options passed to # the field macro. description = maybe_upcase(opts[:upcase], opts[:description] || &quot;&quot;) quote do # We define a clause matching the field name returning its optional # description. def field_description(unquote(name)), do: unquote(description) end end defp maybe_upcase(true, description), do: String.upcase(description) defp maybe_upcase(_, description), do: description # The after_definition callback is called after the struct and its type # have been defined, at the end of the `typedstruct` block. @impl true @spec after_definition(opts :: keyword()) :: Macro.t() def after_definition(_opts) do quote do # Here we just clean the @upcase attribute so that it does not # pollute our user’s modules. Module.delete_attribute(__MODULE__, :upcase) end end end","ref":"TypedStruct.Plugin.html","title":"TypedStruct.Plugin","type":"behaviour"},{"doc":"Injects code after the struct and its type have been defined.","ref":"TypedStruct.Plugin.html#c:after_definition/1","title":"TypedStruct.Plugin.after_definition/1","type":"callback"},{"doc":"Injects code after each field definition.name and type are the exact values passed to the TypedStruct.field/3 macro in the typedstruct block. opts is the concatenation of the options passed to the field macro and those from the plugin init.","ref":"TypedStruct.Plugin.html#c:field/3","title":"TypedStruct.Plugin.field/3","type":"callback"},{"doc":"Injects code where TypedStruct.plugin/2 is called.","ref":"TypedStruct.Plugin.html#c:init/1","title":"TypedStruct.Plugin.init/1","type":"macrocallback"}]